import json
from typing import List, Tuple, Type

import pandas as pd
from scipy import sparse as sps

from irspack.dataset.movielens import MovieLens1MDataManager
from irspack.evaluator import EvaluatorWithColdUser
from irspack.optimizers import (  # SLIMOptimizer,; MultVAEOptimizer,
    AsymmetricCosineKNNOptimizer,
    BaseOptimizer,
    CosineKNNOptimizer,
    DenseSLIMOptimizer,
    IALSOptimizer,
    P3alphaOptimizer,
    RP3betaOptimizer,
    TopPopOptimizer,
    TverskyIndexKNNOptimizer,
)
from irspack.split import split_dataframe_partial_user_holdout

if __name__ == "__main__":

    BASE_CUTOFF = 20

    data_manager = MovieLens1MDataManager()
    df_all = data_manager.read_interaction()

    data_all, _ = split_dataframe_partial_user_holdout(
        df_all,
        "userId",
        "movieId",
        test_user_ratio=0.2,
        val_user_ratio=0.2,
        heldout_ratio_test=0.5,
        heldout_ratio_val=0.5,
    )

    data_train = data_all["train"]
    data_val = data_all["val"]
    data_test = data_all["test"]

    X_train_val_all: sps.csr_matrix = sps.vstack(
        [data_train.X_all, data_val.X_all], format="csr"
    )
    valid_evaluator = EvaluatorWithColdUser(
        input_interaction=data_val.X_train,
        ground_truth=data_val.X_test,
        cutoff=BASE_CUTOFF,
    )
    test_evaluator = EvaluatorWithColdUser(
        input_interaction=data_test.X_train,
        ground_truth=data_test.X_test,
        cutoff=BASE_CUTOFF,
    )

    test_results = []
    validation_results = []

    test_configs: List[Tuple[Type[BaseOptimizer], int]] = [
        (TopPopOptimizer, 1),
        (CosineKNNOptimizer, 40),
        (AsymmetricCosineKNNOptimizer, 40),
        (TverskyIndexKNNOptimizer, 40),
        (P3alphaOptimizer, 40),
        (RP3betaOptimizer, 40),
        (IALSOptimizer, 40),
        (DenseSLIMOptimizer, 20),
        # (SLIMOptimizer, 40), time consuming
        # (MultVAEOptimizer, 5),
    ]
    for optimizer_class, n_trials in test_configs:
        recommender_name = optimizer_class.recommender_class.__name__
        optimizer: BaseOptimizer = optimizer_class(data_train.X_all, valid_evaluator)
        (best_param, validation_result_df) = optimizer.optimize(
            timeout=14400, n_trials=n_trials
        )
        validation_result_df["recommender_name"] = recommender_name
        validation_results.append(validation_result_df)
        pd.concat(validation_results).to_csv(f"validation_scores.csv")
        test_recommender = optimizer.recommender_class(X_train_val_all, **best_param)
        test_recommender.learn()
        test_scores = test_evaluator.get_scores(test_recommender, [5, 10, 20])

        test_results.append(
            dict(name=recommender_name, best_param=best_param, **test_scores)
        )
        with open("test_results.json", "w") as ofs:
            json.dump(test_results, ofs, indent=2)



from .base import BaseUserColdStartRecommender
from .cb_knn import UserCBCosineKNNRecommender
from .linear import LinearMethodRecommender
from .popular import TopPopularRecommender

__all__ = [
    "BaseUserColdStartRecommender",
    "UserCBCosineKNNRecommender",
    "LinearMethodRecommender",
    "TopPopularRecommender",
]



import gc

import numpy as np
from scipy import linalg

from ..definitions import InteractionMatrix
from .base import BaseSimilarityRecommender


class DenseSLIMRecommender(BaseSimilarityRecommender):
    """Implementation of DenseSLIM or Embarrassingly Shallow AutoEncoder (EASE ^R).

    See:

        - `Embarrassingly Shallow Autoencoders for Sparse Data
          <https://arxiv.org/abs/1905.03375>`_

    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.

        reg (float, optional):
            The regularization parameter for ease. Defaults to 1.0.
    """

    def __init__(self, X_train_all: InteractionMatrix, reg: float = 1):

        super(DenseSLIMRecommender, self).__init__(X_train_all)
        self.reg = reg

    def _learn(self) -> None:
        X_train_all_f32 = self.X_train_all.astype(np.float32)
        P = X_train_all_f32.T.dot(X_train_all_f32)
        P_dense: np.ndarray = P.todense()
        del P
        P_dense[np.arange(self.n_items), np.arange(self.n_items)] += self.reg
        gc.collect()
        P_dense = linalg.inv(P_dense, overwrite_a=True)

        gc.collect()
        diag_P_inv = 1 / np.diag(P_dense)
        P_dense *= -diag_P_inv[np.newaxis, :]
        range_ = np.arange(self.n_items)
        P_dense[range_, range_] = 0
        self.W_ = P_dense



import pickle
from typing import List, Type

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.evaluator import Evaluator
from irspack.recommenders import (
    AsymmetricCosineKNNRecommender,
    AsymmetricCosineUserKNNRecommender,
    BaseRecommender,
    CosineKNNRecommender,
    CosineUserKNNRecommender,
    DenseSLIMRecommender,
    IALSRecommender,
    JaccardKNNRecommender,
    NMFRecommender,
    P3alphaRecommender,
    RP3betaRecommender,
    SLIMRecommender,
    TopPopRecommender,
    TruncatedSVDRecommender,
    TverskyIndexKNNRecommender,
)

X_train = sps.csr_matrix(
    np.asfarray([[1, 1, 2, 0, 0], [0, 1, 0, 1, 0], [0, 0, 1, 0, 0], [0, 0, 0, 0, 0]])
)

X_test = sps.csr_matrix(
    np.asfarray([[0, 0, 0, 1, 1], [1, 0, 0, 0, 0], [0, 0, 0, 0, 0], [1, 0, 0, 0, 0]])
)

rec_classes: List[Type[BaseRecommender]] = [
    TopPopRecommender,
    CosineKNNRecommender,
    AsymmetricCosineKNNRecommender,
    TverskyIndexKNNRecommender,
    JaccardKNNRecommender,
    CosineUserKNNRecommender,
    AsymmetricCosineUserKNNRecommender,
    P3alphaRecommender,
    RP3betaRecommender,
    TruncatedSVDRecommender,
    NMFRecommender,
    IALSRecommender,
    DenseSLIMRecommender,
    SLIMRecommender,
]
try:
    from irspack.recommenders.bpr import BPRFMRecommender

    rec_classes.append(BPRFMRecommender)
except:
    pass

try:
    from irspack.recommenders.multvae import MultVAERecommender

    rec_classes.append(MultVAERecommender)
except:
    pass


@pytest.mark.parametrize("RecommenderClass", rec_classes)
def test_recs(RecommenderClass: Type[BaseRecommender]) -> None:
    """Test the learning of recommenders exit normally, and they are picklable.

    Args:
        RecommenderClass (Type[BaseRecommender]): The recommender class to be tested.
    """
    rec = RecommenderClass(X_train)
    rec.learn()

    scores = rec.get_score(np.arange(X_train.shape[0]))
    eval = Evaluator(X_test, 0, 20)
    with pytest.raises(ValueError):
        eval.get_score(rec)
    metrics = eval.get_scores(rec, cutoffs=[X_train.shape[1]])
    assert np.all(np.isfinite(scores))
    assert np.all(~np.isnan(scores))
    for value in metrics.values():
        assert ~np.isnan(value)
        assert np.isfinite(value)
    with open("temp.pkl", "wb") as ofs:
        pickle.dump(rec, ofs)
    with open("temp.pkl", "rb") as ifs:
        rec_dumped: BaseRecommender = pickle.load(ifs)
    score_from_dumped = rec_dumped.get_score(np.arange(X_train.shape[0]))
    np.testing.assert_allclose(scores, score_from_dumped)



from abc import ABC, abstractmethod
from io import BytesIO
from typing import IO, Any, Optional, Type

from fastprogress import progress_bar
from optuna import Trial, exceptions

from irspack.definitions import InteractionMatrix
from irspack.evaluator.evaluator import Evaluator
from irspack.recommenders.base import BaseRecommender


class TrainerBase(ABC):
    """The trainer class for early-stoppable recommenders.
    The training logic of such recommenders (run-epoch, get-score) will be implemtented in the trainers."""

    def __init__(self, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def load_state(self, ifs: IO) -> None:
        """Load past state (with best validation score).

        Args:
            ifs (IO): The stream from which past state will be restored.
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def save_state(self, ofs: IO) -> None:
        """Save the current state into stream.

        Args:
            ofs (IO): the stream to which current state will be saved.
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def run_epoch(self) -> None:
        """Run single epoch."""
        raise NotImplementedError()  # pragma: no cover


class BaseRecommenderWithEarlyStopping(BaseRecommender):
    """The base class for all the early-stoppable recommenders.

    Args:
        X_train_all (csr_matrix|csc_matrix): The train interaction matrix.
        max_epoch (int, optional): The maximal number of epochs to be run. Defaults to 512.
        validate_epoch (int, optional): Frequency of validation score measurement (if any). Defaults to 5.
        score_degradation_max (int, optional): Maximal number of allowed score degradation. Defaults to 5.
    """

    trainer_class: Type[TrainerBase]

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        max_epoch: int = 512,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
        **kwargs: Any,
    ):

        super().__init__(X_train_all, **kwargs)
        self.max_epoch = max_epoch
        self.validate_epoch = validate_epoch
        if max_epoch < validate_epoch:
            raise ValueError("max_epoch must be greater than validate_epoch.")
        self.score_degradation_max = score_degradation_max
        self.trainer: Optional[TrainerBase] = None
        self.best_state: Optional[bytes] = None

    @abstractmethod
    def _create_trainer(self) -> TrainerBase:
        pass  # pragma: no cover

    def start_learning(self) -> None:
        self.trainer = self._create_trainer()

    def run_epoch(self) -> None:
        if self.trainer is None:
            raise RuntimeError("'run_epoch' called before initializing the trainer.")
        self.trainer.run_epoch()

    def save_state(self) -> None:
        if self.trainer is None:
            raise RuntimeError("'save_state' called before initializing the trainer.")
        with BytesIO() as ofs:
            self.trainer.save_state(ofs)
            self.best_state = ofs.getvalue()

    def load_state(self) -> None:
        if self.trainer is None:
            raise RuntimeError("'load_state' called before initializing the trainer.")
        if self.best_state is None:
            raise RuntimeError("'load_state' called before achieving any results.")
        with BytesIO(self.best_state) as ifs:
            self.trainer.load_state(ifs)

    def _learn(self) -> None:
        self.learn_with_optimizer(None, None)

    def learn_with_optimizer(
        self, evaluator: Optional[Evaluator], trial: Optional[Trial]
    ) -> None:
        self.start_learning()
        best_score = -float("inf")
        n_score_degradation = 0
        pb = progress_bar(range(self.max_epoch))
        for epoch in pb:
            self.run_epoch()
            if (epoch + 1) % self.validate_epoch:
                continue

            if evaluator is None:
                continue

            target_score = evaluator.get_target_score(self)

            pb.comment = f"{evaluator.target_metric_name}={target_score}"
            relevant_score = target_score
            if relevant_score > best_score:
                best_score = relevant_score
                self.save_state()
                self.learnt_config["max_epoch"] = epoch + 1
                n_score_degradation = 0
            else:
                n_score_degradation += 1
                if n_score_degradation >= self.score_degradation_max:
                    break
            if trial is not None:
                trial.report(-relevant_score, epoch)
                if trial.should_prune():
                    raise exceptions.TrialPruned()

        if evaluator is not None:
            self.load_state()



from typing import Dict

import numpy as np
import pytest
import scipy.sparse as sps


@pytest.fixture()
def test_interaction_data() -> Dict[str, sps.csr_matrix]:
    X_small = sps.csr_matrix(
        np.asfarray(
            [[1, 1, 2, 3, 4], [0, 1, 0, 1, 0], [0, 0, 1, 0, 0], [0, 0, 0, 0, 0]]
        )
    )
    return {"X_small": X_small}



import uuid

import numpy as np
import pandas as pd

from irspack.split import split_last_n_interaction_df

RNS = np.random.RandomState(0)
n_users = 1000
n_items = 512
df_size = 10000

user_ids = np.asarray([str(uuid.uuid1()) for _ in range(n_users)])
item_ids = np.asarray([str(uuid.uuid1()) for _ in range(n_items)])

df_master = pd.DataFrame(
    dict(
        user_id=RNS.choice(user_ids, size=df_size, replace=True),
        item_id=RNS.choice(item_ids, size=df_size, replace=True),
        ts=RNS.randint(0, 100, size=df_size).astype("datetime64[D]"),
    )
).drop_duplicates()


def test_holdout_fixed_n() -> None:
    df = df_master.copy()
    df_train, df_val = split_last_n_interaction_df(df, "user_id", "ts", 2)
    assert pd.concat([df_train, df_val]).merge(df).shape[0] == df_master.shape[0]
    uid_val_unique = np.unique(df_val.user_id)
    for uid in uid_val_unique:
        interactions_in_val = df_val[df_val.user_id == uid]
        if interactions_in_val.shape[0] == 0:
            continue
        assert interactions_in_val.shape[0] <= 2
        interactions_in_train = df_train[df_train.user_id == uid]
        if interactions_in_train.shape[0] == 0:
            continue
        assert interactions_in_train.ts.max() <= interactions_in_val.ts.min()


def test_holdout_fixed_percentage() -> None:
    df = df_master.copy()
    df_train, df_val = split_last_n_interaction_df(
        df, "user_id", "ts", heldout_ratio=0.5
    )
    assert pd.concat([df_train, df_val]).merge(df).shape[0] == df_master.shape[0]

    uid_val_unique = np.unique(df_val.user_id)
    uid_cnt = df_master.user_id.value_counts()
    for uid in uid_val_unique:
        interactions_in_val = df_val[df_val.user_id == uid]
        if interactions_in_val.shape[0] == 0:
            continue
        # assert interactions_in_val.shape[0] <= (0.5 * uid_cnt.loc[uid])
        interactions_in_train = df_train[df_train.user_id == uid]
        if interactions_in_train.shape[0] == 0:
            continue
        assert interactions_in_train.ts.max() <= interactions_in_val.ts.min()



import json
from typing import Any, Dict, List, Tuple, Type

import pandas as pd
from scipy import sparse as sps

from irspack.dataset.movielens import MovieLens20MDataManager
from irspack.evaluator import EvaluatorWithColdUser
from irspack.optimizers import (
    AsymmetricCosineKNNOptimizer,
    BaseOptimizer,
    CosineKNNOptimizer,
    DenseSLIMOptimizer,
    IALSOptimizer,
    MultVAEOptimizer,
    P3alphaOptimizer,
    RP3betaOptimizer,
    SLIMOptimizer,
    TopPopOptimizer,
)
from irspack.split import split_dataframe_partial_user_holdout

if __name__ == "__main__":

    BASE_CUTOFF = 100

    # We follow the preprocessing of Mult-VAE implementation (https://github.com/dawenl/vae_cf)
    data_manager = MovieLens20MDataManager()
    df_all = data_manager.read_interaction()
    df_all = df_all[df_all.rating >= 4]
    user_cnt = df_all.userId.value_counts()
    user_cnt = user_cnt[user_cnt >= 5]
    df_all = df_all[df_all.userId.isin(user_cnt.index)]

    data_all, _ = split_dataframe_partial_user_holdout(
        df_all,
        "userId",
        "movieId",
        n_test_user=10000,
        n_val_user=10000,
        heldout_ratio_val=0.2,
        heldout_ratio_test=0.2,
    )

    data_train = data_all["train"]
    data_val = data_all["val"]
    data_test = data_all["test"]

    X_train_val_all: sps.csr_matrix = sps.vstack(
        [data_train.X_all, data_val.X_all], format="csr"
    )
    valid_evaluator = EvaluatorWithColdUser(
        input_interaction=data_val.X_train,
        ground_truth=data_val.X_test,
        cutoff=BASE_CUTOFF,
    )
    test_evaluator = EvaluatorWithColdUser(
        input_interaction=data_test.X_train,
        ground_truth=data_test.X_test,
        cutoff=BASE_CUTOFF,
    )

    test_results = []
    validation_results = []

    test_configs: List[Tuple[Type[BaseOptimizer], int, Dict[str, Any]]] = [
        (TopPopOptimizer, 1, dict()),
        (CosineKNNOptimizer, 40, dict()),
        (AsymmetricCosineKNNOptimizer, 40, dict()),
        (P3alphaOptimizer, 30, dict(alpha=1)),
        (RP3betaOptimizer, 40, dict(alpha=1)),
        (IALSOptimizer, 40, dict()),
        (DenseSLIMOptimizer, 20, dict()),
        (
            MultVAEOptimizer,
            1,
            dict(
                dim_z=200, enc_hidden_dims=600, kl_anneal_goal=0.2
            ),  # nothing to tune, use the parameters used in the paper.
        ),
        (SLIMOptimizer, 40, dict()),  # Note: this is a heavy one.
    ]
    for optimizer_class, n_trials, config in test_configs:
        recommender_name = optimizer_class.recommender_class.__name__
        optimizer: BaseOptimizer = optimizer_class(
            data_train.X_all,
            valid_evaluator,
            fixed_params=config,
        )
        (best_param, validation_result_df) = optimizer.optimize(n_trials=n_trials)
        validation_result_df["recommender_name"] = recommender_name
        validation_results.append(validation_result_df)
        pd.concat(validation_results).to_csv(f"validation_scores.csv")
        test_recommender = optimizer.recommender_class(X_train_val_all, **best_param)
        test_recommender.learn()
        test_scores = test_evaluator.get_scores(test_recommender, [20, 50, 100])

        test_results.append(
            dict(name=recommender_name, best_param=best_param, **test_scores)
        )
        with open("test_results.json", "w") as ofs:
            json.dump(test_results, ofs, indent=2)



from typing import Dict, Tuple

import numpy as np
import pytest
import scipy.sparse as sps
from scipy.optimize import minimize

from irspack.recommenders import IALSRecommender


def test_ials_overfit_cholesky(
    test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:
    X = test_interaction_data["X_small"]
    rec = IALSRecommender(
        X,
        n_components=4,
        alpha=0,
        reg=0.001,
        use_cg=False,
    )
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.compute_user_embedding(X)
    ivec = rec.compute_item_embedding(X)
    X = X.toarray()
    X[X.nonzero()] = 1.0
    reprod = uvec.dot(ivec.T)
    np.testing.assert_allclose(reprod, X, rtol=1e-2, atol=1e-2)


def test_ials_overfit_cg(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    X = test_interaction_data["X_small"]
    rec = IALSRecommender(
        X, n_components=4, alpha=0, reg=1e-2, use_cg=True, max_cg_steps=4
    )
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.trainer.core_trainer.transform_user(X.tocsr().astype(np.float32))
    ivec = rec.trainer.core_trainer.transform_item(X.tocsr().astype(np.float32))
    X_dense = X.toarray()
    X_dense[X_dense.nonzero()] = 1.0
    reproduced_user_vector = uvec.dot(ivec.T)
    np.testing.assert_allclose(reproduced_user_vector, X_dense, rtol=1e-2, atol=1e-2)

    X_reproduced = rec.get_score_cold_user(X)
    np.testing.assert_allclose(X_reproduced, X_dense, rtol=1e-2, atol=1e-2)

    with pytest.raises(ValueError):
        _ = rec.compute_item_embedding(X.T)

    reproduced_item_vector = rec.compute_item_embedding(X)
    X_reproduced_item = rec.get_score_from_item_embedding(
        np.arange(X.shape[0]), reproduced_item_vector
    )
    np.testing.assert_allclose(X_reproduced_item, X_dense, rtol=1e-2, atol=1e-2)


@pytest.mark.xfail
def test_ials_cg_underfit(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    X = test_interaction_data["X_small"]
    rec = IALSRecommender(
        X,
        n_components=4,
        alpha=0,
        reg=1e-2,
        use_cg=True,
        max_cg_steps=1,
        max_epoch=10,
    )
    with pytest.raises(RuntimeError):
        _ = rec.core_trainer.user
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.core_trainer.user
    ivec = rec.core_trainer.item
    X_dense = X.toarray()
    reprod = uvec.dot(ivec.T)
    np.testing.assert_allclose(reprod, X_dense, rtol=1e-2, atol=1e-2)


def test_ials_overfit_nonzero_alpha(
    test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:
    ALPHA = 4.5
    REG = 3
    X = test_interaction_data["X_small"]
    rec_chol = IALSRecommender(
        X, n_components=4, alpha=ALPHA, reg=REG, use_cg=False, max_epoch=5
    )
    rec_chol.learn()
    assert rec_chol.trainer is not None
    uvec_chol = rec_chol.trainer.core_trainer.transform_user(
        X.tocsr().astype(np.float32)
    )
    ivec_chol = rec_chol.trainer.core_trainer.transform_item(
        X.tocsr().astype(np.float32)
    )

    rec_cg = IALSRecommender(
        X,
        n_components=4,
        alpha=ALPHA,
        reg=REG,
        use_cg=True,
        max_cg_steps=4,
        max_epoch=5,
    )
    rec_cg.learn()
    assert rec_cg.trainer is not None
    uvec_cg = rec_cg.trainer.core_trainer.transform_user(X.tocsr())
    ivec_cg = rec_cg.trainer.core_trainer.transform_item(X.tocsr())

    np.testing.assert_allclose(uvec_chol, uvec_cg, atol=1e-3, rtol=1e-4)
    np.testing.assert_allclose(ivec_chol, ivec_cg, atol=1e-3, rtol=1e-4)


def ials_grad(
    X: sps.csr_matrix,
    u: np.ndarray,
    v: np.ndarray,
    reg: float,
    alpha: float,
    epsilon: float,
) -> Tuple[np.ndarray, np.ndarray]:
    nu = u.shape[0]
    ni = v.shape[0]

    uv = u.dot(v.T)
    result_u = np.zeros_like(u)
    result_v = np.zeros_like(v)
    for uind in range(nu):
        for iind in range(ni):
            x = X[uind, iind]
            if x == 0:
                sc = uv[uind, iind]
            else:
                sc = (1 + alpha * np.log(1 + x / epsilon)) * (uv[uind, iind] - 1)

            result_u[uind, :] += v[iind] * sc
            result_v[iind, :] += u[uind] * sc
    result_u += reg * u
    result_v += reg * v
    return result_u, result_v


def test_ials_overfit_cholesky_logscale(
    test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:

    ALPHA = 1.0
    REG = 0.1
    EPSILON = 1.0
    N_COMPONENTS = 1
    X = test_interaction_data["X_small"]
    rec_chol = IALSRecommender(
        X,
        n_components=N_COMPONENTS,
        alpha=ALPHA,
        reg=REG,
        use_cg=False,
        epsilon=EPSILON,
        confidence_scaling="log",
        max_epoch=200,
        init_std=1e-1,
    )
    rec_chol.learn()
    assert rec_chol.trainer is not None
    uvec_chol = rec_chol.get_user_embedding()
    ivec_chol_cold = rec_chol.compute_item_embedding(X)
    print(X.shape)

    grad_uvec_chol, grad_ivec_chol = ials_grad(
        X, uvec_chol, ivec_chol_cold, REG, ALPHA, EPSILON
    )

    np.testing.assert_allclose(grad_ivec_chol, np.zeros_like(grad_ivec_chol), atol=1e-5)
    np.testing.assert_allclose(grad_uvec_chol, np.zeros_like(grad_uvec_chol), atol=1e-5)



import logging
from typing import Any, Dict, List, Optional, Type

import optuna
from sklearn.model_selection import train_test_split

from irspack.definitions import InteractionMatrix
from irspack.parameter_tuning import Suggestion, overwrite_suggestions
from irspack.user_cold_start.evaluator import UserColdStartEvaluator
from irspack.user_cold_start.recommenders.base import (
    BaseUserColdStartRecommender,
    ProfileMatrix,
)
from irspack.utils.default_logger import get_default_logger


class BaseOptimizer:
    default_tune_range: List[Suggestion]
    recommender_class: Type[BaseUserColdStartRecommender]

    def __init__(
        self,
        X_train: InteractionMatrix,
        profile_train: ProfileMatrix,
        evaluator: UserColdStartEvaluator,
        target_metric: str = "ndcg",
        suggest_overwrite: List[Suggestion] = list(),
        fixed_params: Dict[str, Any] = dict(),
        logger: Optional[logging.Logger] = None,
    ):
        if logger is None:
            logger = get_default_logger()

        self.logger = logger

        self.X_train = X_train
        self.profile_train = profile_train
        self.evaluator = evaluator
        self.target_metric = target_metric
        self.current_trial: int = 0
        self.best_trial_index: Optional[int] = None
        self.best_val = float("inf")
        self.best_params: Optional[Dict[str, Any]] = None
        self.learnt_config_best: Dict[str, Any] = dict()

        self.valid_results: List[Dict[str, float]] = []
        self.tried_configs: List[Dict[str, Any]] = []
        self.suggestions = overwrite_suggestions(
            self.default_tune_range, suggest_overwrite, fixed_params
        )
        self.fixed_params = fixed_params

    def _suggest(self, trial: optuna.Trial) -> Dict[str, Any]:
        parameters: Dict[str, Any] = dict()
        for s in self.suggestions:
            parameters[s.name] = s.suggest(trial)
        return parameters

    def optimize(
        self,
        n_trials: int = 20,
        timeout: Optional[int] = None,
        random_seed: Optional[int] = None,
    ) -> Dict[str, Any]:
        def objective(trial: optuna.Trial) -> float:
            param_dict = dict(**self._suggest(trial), **self.fixed_params)
            recommender = self.recommender_class(
                self.X_train, self.profile_train, **param_dict
            )
            recommender.learn()
            val_score: float = self.evaluator.get_score(recommender)[self.target_metric]
            if (-val_score) < self.best_val:
                self.best_val = -val_score
                self.best_params = param_dict
                self.logger.info("Found best %s using this config.", self.target_metric)
                self.best_trial_index = self.current_trial

            return -val_score

        study = optuna.create_study(
            sampler=optuna.samplers.TPESampler(seed=random_seed)
        )
        study.optimize(objective, n_trials, timeout=timeout)
        if self.best_params is None:
            raise RuntimeError(
                "best parameter not found (possibly because no trial has been made)"
            )
        best_params: Dict[str, Any] = self.best_params
        return best_params

    @classmethod
    def split_and_optimize(
        cls,
        X_train: InteractionMatrix,
        X_profile: ProfileMatrix,
        n_trials: int = 20,
        target_metric: str = "ndcg",
        split_config: Dict[str, Any] = dict(test_size=0.2, random_state=42),
        evaluator_config: Dict[str, Any] = dict(),
        timeout: Optional[int] = None,
        suggest_overwrite: List[Suggestion] = [],
        fixed_params: Dict[str, Any] = dict(),
        logger: Optional[logging.Logger] = None,
        random_seed: Optional[int] = None,
    ) -> Dict[str, Any]:
        X_tt, X_tv, profile_tt, profile_tv = train_test_split(
            X_train, X_profile, **split_config
        )
        X_tt.sort_indices()
        X_tv.sort_indices()
        profile_tt.sort_indices()
        profile_tv.sort_indices()
        evaluator = UserColdStartEvaluator(X_tv, profile_tv, **evaluator_config)
        optimizer = cls(
            X_tt,
            profile_tt,
            evaluator,
            target_metric=target_metric,
            suggest_overwrite=suggest_overwrite,
            fixed_params=fixed_params,
            logger=logger,
        )
        return optimizer.optimize(
            n_trials=n_trials, timeout=timeout, random_seed=random_seed
        )



import os
from typing import Optional


def get_n_threads(n_threads: Optional[int]) -> int:
    if n_threads is not None:
        return n_threads
    else:
        try:
            n_threads_cand = os.environ.get(
                "IRSPACK_NUM_THREADS_DEFAULT", os.cpu_count()
            )
            return int(n_threads_cand or 1)
        except:
            raise ValueError(
                'failed to interpret "IRSPACK_NUM_THREADS_DEFAULT" as an integer.'
            )



m: int
n: int
from numpy import float32

import irspack.utils._util_cpp
from typing import *
from typing import Iterable as iterable
from typing import Iterator as iterator
from numpy import float64

_Shape = Tuple[int, ...]
import numpy
import scipy.sparse

__all__ = [
    "okapi_BM_25_weight",
    "remove_diagonal",
    "retrieve_recommend_from_score",
    "rowwise_train_test_split_by_fixed_n",
    "rowwise_train_test_split_by_ratio",
    "slim_weight_allow_negative",
    "slim_weight_positive_only",
    "sparse_mm_threaded",
    "tf_idf_weight",
]

def okapi_BM_25_weight(
    X: scipy.sparse.csr_matrix[float64], k1: float = 1.2, b: float = 0.75
) -> scipy.sparse.csr_matrix[float64]:
    pass

def remove_diagonal(
    arg0: scipy.sparse.csr_matrix[float64],
) -> scipy.sparse.csr_matrix[float64]:
    pass

def retrieve_recommend_from_score(
    score: numpy.ndarray[float64, _Shape[m, n]],
    allowed_indices: List[List[int]],
    cutoff: int,
    n_threads: int = 1,
) -> List[List[Tuple[int, float]]]:
    pass

def rowwise_train_test_split_by_fixed_n(
    arg0: scipy.sparse.csr_matrix[float64], arg1: int, arg2: int
) -> Tuple[scipy.sparse.csr_matrix[float64], scipy.sparse.csr_matrix[float64]]:
    pass

def rowwise_train_test_split_by_ratio(
    arg0: scipy.sparse.csr_matrix[float64], arg1: int, arg2: float, arg3: bool
) -> Tuple[scipy.sparse.csr_matrix[float64], scipy.sparse.csr_matrix[float64]]:
    pass

def slim_weight_allow_negative(
    X: scipy.sparse.csr_matrix[float32],
    n_threads: int,
    n_iter: int,
    l2_coeff: float,
    l1_coeff: float,
    tol: float,
    top_k: int = -1,
) -> scipy.sparse.csc_matrix[float32]:
    pass

def slim_weight_positive_only(
    X: scipy.sparse.csr_matrix[float32],
    n_threads: int,
    n_iter: int,
    l2_coeff: float,
    l1_coeff: float,
    tol: float,
    top_k: int = -1,
) -> scipy.sparse.csc_matrix[float32]:
    pass

def sparse_mm_threaded(
    arg0: scipy.sparse.csr_matrix[float64],
    arg1: scipy.sparse.csc_matrix[float64],
    arg2: int,
) -> numpy.ndarray[float64, _Shape[m, n]]:
    pass

def tf_idf_weight(
    X: scipy.sparse.csr_matrix[float64], smooth: bool = True
) -> scipy.sparse.csr_matrix[float64]:
    pass



import os

import pandas as pd

from .base import BaseMovieLenstDataLoader


class MovieLens20MDataManager(BaseMovieLenstDataLoader):
    DOWNLOAD_URL = "http://files.grouplens.org/datasets/movielens/ml-20m.zip"
    DEFAULT_PATH = os.path.expanduser("~/.ml-20m.zip")
    INTERACTION_PATH = "ml-20m/ratings.csv"

    def read_interaction(self) -> pd.DataFrame:
        with self._read_as_istream(self.INTERACTION_PATH) as ifs:
            df = pd.read_csv(ifs)
            df["timestamp"] = pd.to_datetime(df["timestamp"], unit="s")
            return df



from abc import ABC, abstractmethod
from typing import Any, List

from irspack.definitions import DenseScoreArray, InteractionMatrix, ProfileMatrix
from irspack.parameter_tuning import Suggestion


class BaseUserColdStartRecommender(ABC):
    default_tune_range: List[Suggestion] = []

    def __init__(
        self, X_interaction: InteractionMatrix, X_profile: ProfileMatrix, **kwargs: Any
    ):
        assert X_interaction.shape[0] == X_profile.shape[0]
        self.n_users = X_profile.shape[0]
        self.n_items = X_interaction.shape[1]
        self.profile_dimension = X_profile.shape[1]
        self.X_profile = X_profile
        self.X_interaction = X_interaction

    @abstractmethod
    def _learn(self) -> None:
        pass

    def learn(self) -> "BaseUserColdStartRecommender":
        self._learn()
        return self

    @abstractmethod
    def get_score(self, profile: ProfileMatrix) -> DenseScoreArray:
        raise NotImplementedError("implemented in the descendant")



import uuid
from typing import Any, Dict, List, Tuple

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.definitions import DenseScoreArray, InteractionMatrix
from irspack.recommenders import BaseRecommender
from irspack.utils.id_mapping import IDMappedRecommender


class MockRecommender(BaseRecommender):
    def __init__(self, X_all: sps.csr_matrix, scores: np.ndarray) -> None:
        super().__init__(X_all)
        self.scores = scores

    def get_score(self, user_indices: np.ndarray) -> np.ndarray:
        return self.scores[user_indices]

    def _learn(self) -> None:
        pass

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        score = np.exp(X.toarray())
        score /= score.sum(axis=1)[:, None]
        return score


def check_descending(id_score: List[Tuple[Any, float]]) -> None:
    for i, (_, score) in enumerate(id_score):
        if i > 0:
            if score > id_score[i - 1][1]:
                raise RuntimeError("not descending!")


def test_check_descending_1() -> None:
    check_descending([("1", 100), ("2", 99), ("100", 99), ("3", -1000)])
    with pytest.raises(RuntimeError):
        check_descending([("1", 100), ("2", 99), ("3", -1000), ("", -999)])


RNS = np.random.RandomState(0)
n_users = 31
n_items = 42
user_ids = [str(uuid.uuid4()) for _ in range(n_users)]
item_ids = [str(uuid.uuid4()) for _ in range(n_items)]
item_id_set = set(item_ids)
score = RNS.randn(n_users, n_items)
X = sps.csr_matrix((score + RNS.randn(*score.shape)) > 0).astype(np.float64)
rec = MockRecommender(X, score).learn()


def test_basic_usecase() -> None:

    with pytest.raises(ValueError):
        mapped_rec = IDMappedRecommender(rec, user_ids, item_ids + [str(uuid.uuid4())])

    mapped_rec = IDMappedRecommender(rec, user_ids, item_ids)

    with pytest.raises(RuntimeError):
        mapped_rec.get_recommendation_for_known_user_id(str(uuid.uuid4()))

    for i, uid in enumerate(user_ids):
        nonzero_items = [item_ids[j] for j in X[i].nonzero()[1]]
        recommended = mapped_rec.get_recommendation_for_known_user_id(
            uid, cutoff=n_items
        )
        check_descending(recommended)
        recommended_ids = {rec[0] for rec in recommended}
        assert len(recommended_ids.intersection(nonzero_items)) == 0
        assert len(recommended_ids.union(nonzero_items)) == n_items
        cutoff = n_items // 2
        recommended_with_cutoff = mapped_rec.get_recommendation_for_known_user_id(
            uid, cutoff=cutoff
        )
        check_descending(recommended_with_cutoff)
        assert len(recommended_with_cutoff) <= cutoff

        forbbiden_item = list(
            set(
                RNS.choice(
                    list(item_id_set.difference(nonzero_items)),
                    replace=False,
                    size=(n_items - len(nonzero_items)) // 2,
                )
            )
        )

        recommended_with_restriction = mapped_rec.get_recommendation_for_known_user_id(
            uid, cutoff=n_items, forbidden_item_ids=forbbiden_item
        )
        check_descending(recommended_with_restriction)
        for iid, score in recommended_with_restriction:
            assert iid not in forbbiden_item

        random_allowed_items = list(
            RNS.choice(
                list(item_id_set.difference(nonzero_items)),
                size=min(n_items - len(nonzero_items), n_items // 3),
            )
        )
        random_allowed_items += [str(uuid.uuid1())]
        with_allowed_item = mapped_rec.get_recommendation_for_known_user_id(
            uid,
            cutoff=n_items,
            allowed_item_ids=random_allowed_items,
        )
        check_descending(with_allowed_item)
        for id, _ in with_allowed_item:
            assert id in random_allowed_items

        allowed_only_forbidden = mapped_rec.get_recommendation_for_known_user_id(
            uid,
            cutoff=n_items,
            forbidden_item_ids=forbbiden_item,
            allowed_item_ids=forbbiden_item,
        )
        assert not allowed_only_forbidden

        recommendations_for_coldstart = mapped_rec.get_recommendation_for_new_user(
            nonzero_items, cutoff=n_items
        )
        coldstart_rec_results = {_[0] for _ in recommendations_for_coldstart}
        assert len(coldstart_rec_results.intersection(nonzero_items)) == 0
        assert len(coldstart_rec_results.union(nonzero_items)) == n_items

    nonzero_batch: List[List[str]] = []
    forbidden_items_batch: List[List[str]] = []
    allowed_items_batch: List[List[str]] = []
    for i, _ in enumerate(user_ids):
        nonzero_items = [item_ids[j] for j in X[i].nonzero()[1]]
        nonzero_batch.append(nonzero_items)
        forbidden_items_batch.append(
            list(
                set(
                    RNS.choice(
                        list(item_id_set.difference(nonzero_items)),
                        replace=False,
                        size=(n_items - len(nonzero_items)) // 2,
                    )
                )
            )
        )
        allowed_items_batch.append(
            list(
                RNS.choice(
                    list(item_id_set.difference(nonzero_items)),
                    size=min(n_items - len(nonzero_items), n_items // 3),
                )
            )
        )
    batch_result_non_masked = mapped_rec.get_recommendation_for_new_user_batch(
        nonzero_batch, cutoff=n_items, n_threads=1
    )

    assert len(batch_result_non_masked) == n_users
    for recommended_using_batch, uid, nonzero_items in zip(
        batch_result_non_masked, user_ids, nonzero_batch
    ):
        check_descending(recommended_using_batch)
        recommended_ids = {rec[0] for rec in recommended_using_batch}
        assert len(recommended_ids.intersection(nonzero_items)) == 0
        assert len(recommended_ids.union(nonzero_items)) == n_items

    batch_result_non_masked_cutoff_3 = mapped_rec.get_recommendation_for_new_user_batch(
        nonzero_batch, cutoff=3, n_threads=2
    )

    assert len(batch_result_non_masked_cutoff_3) == n_users
    for recommended_using_batch_cutoff_3, uid, nonzero_items in zip(
        batch_result_non_masked_cutoff_3, user_ids, nonzero_batch
    ):
        assert len(recommended_using_batch_cutoff_3) == 3
        check_descending(recommended_using_batch_cutoff_3)
        recommended_ids = {rec[0] for rec in recommended_using_batch_cutoff_3}
        for rid in recommended_ids:
            assert rid not in nonzero_items

    batch_result_masked = mapped_rec.get_recommendation_for_new_user_batch(
        nonzero_batch,
        cutoff=n_items,
        forbidden_item_ids=forbidden_items_batch,
        n_threads=1,
    )
    assert len(batch_result_masked) == n_users
    for recommended_using_batch_masked, uid, nonzero_items, forbidden_items in zip(
        batch_result_masked, user_ids, nonzero_batch, forbidden_items_batch
    ):
        check_descending(recommended_using_batch_masked)
        recommended_ids = {rec[0] for rec in recommended_using_batch_masked}
        assert len(recommended_ids.intersection(nonzero_items)) == 0
        assert (
            len(recommended_ids.union(nonzero_items).union(forbidden_items)) == n_items
        )

    batch_result_masked_restricted = mapped_rec.get_recommendation_for_new_user_batch(
        nonzero_batch,
        cutoff=n_items,
        forbidden_item_ids=forbidden_items_batch,
        allowed_item_ids=allowed_items_batch,
        n_threads=1,
    )
    assert len(batch_result_masked_restricted) == n_users
    for (
        recommended_using_batch_masked_restricted,
        uid,
        nonzero_items,
        forbidden_items,
        allowed_items,
    ) in zip(
        batch_result_masked_restricted,
        user_ids,
        nonzero_batch,
        forbidden_items_batch,
        allowed_items_batch,
    ):
        check_descending(recommended_using_batch_masked_restricted)
        recommended_ids = {rec[0] for rec in recommended_using_batch_masked_restricted}
        assert not recommended_ids.intersection(nonzero_items)
        for rid in recommended_ids:
            assert rid in allowed_items
            assert rid not in forbidden_items

    nonzero_batch_dict: List[Dict[str, float]] = []
    for i, _ in enumerate(user_ids):
        profile = {item_ids[j]: 2.0 for j in X[i].nonzero()[1]}
        nonzero_batch_dict.append(profile)
    batch_result_dict = mapped_rec.get_recommendation_for_new_user_batch(
        nonzero_batch_dict, cutoff=n_items, n_threads=1
    )
    for i, result in enumerate(batch_result_dict):
        nnz = len(X[i].nonzero()[1])
        softmax_denom = X.shape[1] - nnz + np.exp(2) * nnz
        for _, score in result:
            assert score == pytest.approx(1 / softmax_denom)



import numpy as np

from irspack.definitions import DenseScoreArray
from irspack.user_cold_start.recommenders.base import (
    BaseUserColdStartRecommender,
    ProfileMatrix,
)


class TopPopularRecommender(BaseUserColdStartRecommender):
    def _learn(self) -> None:
        self.popularity = self.X_interaction.sum(axis=0).astype(np.float64)

    def get_score(self, profile: ProfileMatrix) -> DenseScoreArray:
        target_n_users = profile.shape[0]
        return np.repeat(self.popularity, target_n_users, axis=0)



import enum
import pickle
from typing import IO, Optional

import numpy as np
import scipy.sparse as sps

from irspack.utils import get_n_threads

from ..definitions import (
    DenseMatrix,
    DenseScoreArray,
    InteractionMatrix,
    UserIndexArray,
)
from ._ials import IALSLearningConfigBuilder
from ._ials import IALSTrainer as CoreTrainer
from .base import BaseRecommenderWithItemEmbedding, BaseRecommenderWithUserEmbedding
from .base_earlystop import BaseRecommenderWithEarlyStopping, TrainerBase


class IALSTrainer(TrainerBase):
    def __init__(
        self,
        X: InteractionMatrix,
        n_components: int,
        alpha: float,
        reg: float,
        init_std: float,
        use_cg: bool,
        max_cg_steps: int,
        random_seed: int,
        n_threads: int,
    ):
        X_train_all_f32 = X.astype(np.float32)
        config = (
            IALSLearningConfigBuilder()
            .set_K(n_components)
            .set_init_stdev(init_std)
            .set_alpha(alpha)
            .set_reg(reg)
            .set_n_threads(n_threads)
            .set_use_cg(use_cg)
            .set_max_cg_steps(max_cg_steps)
            .set_random_seed(random_seed)
            .build()
        )

        self.core_trainer = CoreTrainer(config, X_train_all_f32)

    def load_state(self, ifs: IO) -> None:
        params = pickle.load(ifs)
        self.core_trainer.user = params["user"]
        self.core_trainer.item = params["item"]

    def save_state(self, ofs: IO) -> None:
        pickle.dump(
            dict(user=self.core_trainer.user, item=self.core_trainer.item),
            ofs,
            protocol=pickle.HIGHEST_PROTOCOL,
        )

    def run_epoch(self) -> None:
        self.core_trainer.step()


class IALSConfigScaling(enum.Enum):
    none = enum.auto()
    log = enum.auto()


class IALSRecommender(
    BaseRecommenderWithEarlyStopping,
    BaseRecommenderWithUserEmbedding,
    BaseRecommenderWithItemEmbedding,
):
    r"""Implementation of Implicit Alternating Least Squares(IALS) or Weighted Matrix Factorization(WMF).

    It tries to minimize the following loss:

    .. math ::

        \frac{1}{2} \sum _{u, i} c_{ui} (\mathbf{u}_u \cdot \mathbf{v}_i - \mathbb{1}_{r_{ui} > 0}) ^ 2 +
        \frac{\text{reg}}{2} \left( \sum _u || \mathbf{u}_u || ^2 + \sum _i || \mathbf{v}_i || ^2 \right)


    See the seminal paper:

        - `Collaborative filtering for implicit feedback datasets
          <http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.167.5120&rep=rep1&type=pdf>`_



    To speed up the learning procedure, we have also implemented the conjugate gradient descent version following:

        - `Applications of the conjugate gradient method for implicit feedback collaborative filtering
          <https://dl.acm.org/doi/abs/10.1145/2043932.2043987>`_


    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        n_components (int, optional):
            The dimension for latent factor. Defaults to 20.
        alpha (float, optional):
            The confidence parameter alpha in the original paper. Defaults to 0.0.
        reg (float, optional) :
            Regularization coefficient for both user & item factors. Defaults to 1e-3.
        confidence_scaling (str, optional) :
            Specifies how to scale confidence scaling :math:`c_{ui}`. Must be either "none" or "log".
            If "none", the non-zero "rating" :math:`r_{ui}` yields

            .. math ::

                c_{ui} = 1 + \alpha r_{ui}

            If "log",

            .. math ::

                c_{ui} = 1 + \alpha \log (1 + r_{ui} / \epsilon )

            Defaults to "none".
        epsilon (float, optional):
            The :math:`\epsilon` parameter for log-scaling described above.
            Will not have any effect if `confidence_scaling` is "none".
            Defaults to 1.0f.
        init_std (float, optional):
            Standard deviation for initialization normal distribution. Defaults to 0.1.
        use_cg (bool, optional):
            Whether to use the conjugate gradient method. Defaults to True.
        max_cg_steps (int, optional):
            Maximal number of conjute gradient descent steps. Defaults to 3.
            Ignored when ``use_cg=False``. By increasing this parameter, the result will be closer to
            Cholesky decomposition method (i.e., when ``use_cg = False``), but it wll take longer time.
        validate_epoch (int, optional):
            Frequency of validation score measurement (if any). Defaults to 5.
        score_degradation_max (int, optional):
            Maximal number of allowed score degradation. Defaults to 5.
        n_threads (Optional[int], optional):
            Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if there is no such an environment variable, it will be set to ``os.cpu_count()``. Defaults to None.
        max_epoch (int, optional):
            Maximal number of epochs. Defaults to 512.
    """

    @classmethod
    def _scale_X(
        cls, X: sps.csr_matrix, scheme: IALSConfigScaling, epsilon: float
    ) -> sps.csr_matrix:
        if scheme is IALSConfigScaling.none:
            return X
        else:
            X_ret: sps.csr_matrix = X.copy()
            X_ret.data = np.log(1 + X_ret.data / epsilon)
            return X_ret

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        n_components: int = 20,
        alpha: float = 0.0,
        reg: float = 1e-3,
        confidence_scaling: str = "none",
        epsilon: float = 1.0,
        init_std: float = 0.1,
        use_cg: bool = True,
        max_cg_steps: int = 3,
        random_seed: int = 42,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
        n_threads: Optional[int] = None,
        max_epoch: int = 512,
    ) -> None:

        super().__init__(
            X_train_all,
            max_epoch=max_epoch,
            validate_epoch=validate_epoch,
            score_degradation_max=score_degradation_max,
        )

        self.n_components = n_components
        self.alpha = alpha
        self.reg = reg
        self.init_std = init_std
        self.use_cg = use_cg
        self.max_cg_steps = max_cg_steps
        self.confidence_scaling = IALSConfigScaling[confidence_scaling]
        self.epsilon = epsilon
        self.random_seed = random_seed
        self.n_threads = get_n_threads(n_threads)

        self.trainer: Optional[IALSTrainer] = None

    def _create_trainer(self) -> TrainerBase:
        return IALSTrainer(
            self._scale_X(self.X_train_all, self.confidence_scaling, self.epsilon),
            self.n_components,
            self.alpha,
            self.reg,
            self.init_std,
            self.use_cg,
            self.max_cg_steps,
            self.random_seed,
            self.n_threads,
        )

    @property
    def core_trainer(self) -> CoreTrainer:
        if self.trainer is None:
            raise RuntimeError("tried to fetch core_trainer before the training.")
        return self.trainer.core_trainer

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return self.core_trainer.user[user_indices].dot(self.get_item_embedding().T)

    def get_score_block(self, begin: int, end: int) -> DenseScoreArray:
        return self.core_trainer.user_scores(begin, end)

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        user_vector = self.compute_user_embedding(X)
        return self.get_score_from_user_embedding(user_vector)

    def get_user_embedding(self) -> DenseMatrix:
        return self.core_trainer.user.astype(np.float64)

    def get_score_from_user_embedding(
        self, user_embedding: DenseMatrix
    ) -> DenseScoreArray:
        return user_embedding.dot(self.get_item_embedding().T).astype(np.float64)

    def get_item_embedding(self) -> DenseMatrix:
        return self.core_trainer.item.astype(np.float64)

    def compute_user_embedding(self, X: InteractionMatrix) -> DenseMatrix:
        r"""Given an unknown users' interaction with known items,
        computes the latent factors of the users by least square (fixing item embeddings).

        Parameters:
            X:
                The interaction history of the new users.
                ``X.shape[1]`` must be equal to ``self.n_items``.
        """
        return self.core_trainer.transform_user(
            self._scale_X(
                sps.csr_matrix(X).astype(np.float32),
                self.confidence_scaling,
                self.epsilon,
            )
        )

    def compute_item_embedding(self, X: InteractionMatrix) -> DenseMatrix:
        r"""Given an unknown items' interaction with known user,
        computes the latent factors of the items by least square (fixing user embeddings).

        Parameters:
            X:
                The interaction history of the new users.
                ``X.shape[0]`` must be equal to ``self.n_users``.
        """

        return self.core_trainer.transform_item(
            self._scale_X(
                sps.csr_matrix(X).astype(np.float32),
                self.confidence_scaling,
                self.epsilon,
            )
        )

    def get_score_from_item_embedding(
        self, user_indices: UserIndexArray, item_embedding: DenseMatrix
    ) -> DenseScoreArray:
        return (
            self.core_trainer.user[user_indices]
            .dot(item_embedding.T)
            .astype(np.float64)
        )



from dataclasses import dataclass
from functools import partial
from logging import Logger
from typing import Any, Callable, Iterator, List, Optional, Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax
import optuna
from fastprogress import progress_bar
from jax._src.random import PRNGKey
from optuna import exceptions
from scipy import sparse as sps
from sklearn.model_selection import train_test_split

from irspack.utils.default_logger import get_default_logger


@dataclass
class MLP:
    predict_function: Callable[[hk.Params, jnp.ndarray, bool], jnp.ndarray]
    params: hk.Params
    rng_key = PRNGKey(0)

    def predict(self, X: jnp.ndarray) -> np.ndarray:
        f: Callable[[hk.Params, PRNGKey, jnp.ndarray, bool], jnp.ndarray] = getattr(
            self, "predict_function"
        )
        return np.asarray(f(self.params, self.rng_key, X, False), dtype=np.float32)


@dataclass
class MLPTrainingConfig:
    intermediate_dims: List[int]
    dropout: float = 0.0
    weight_decay: float = 0.0
    best_epoch: Optional[int] = None
    activation: Callable[[jnp.ndarray], jnp.ndarray] = jnp.tanh
    learning_rate: float = 1e-3


@dataclass
class MLPSearchConfig:
    n_layers: int = 1
    tune_weight_decay: bool = False
    tune_dropout: bool = True
    layer_dim_max: int = 512

    def suggest(self, trial: optuna.Trial) -> MLPTrainingConfig:
        layer_dims = [
            trial.suggest_int(f"dim_{l}", 1, self.layer_dim_max)
            for l in range(self.n_layers)
        ]
        if not self.tune_dropout:
            dropout = 0.0
        else:
            dropout = trial.suggest_uniform("dropout", 0, 1)

        return MLPTrainingConfig(layer_dims, dropout)


def create_mlp(
    dim_out: int, config: MLPTrainingConfig
) -> Callable[[jnp.ndarray, bool], Any]:
    def mlp_function(X: jnp.ndarray, training: bool) -> Any:
        layers: List[Any] = []
        for d_o in config.intermediate_dims:
            if training:
                layers.append(
                    lambda x: hk.dropout(hk.next_rng_key(), config.dropout, x)
                )
            layers.append(hk.Linear(d_o))
            layers.append(config.activation)
        layers.append(hk.Linear(dim_out))
        return hk.Sequential(layers)(X)

    return mlp_function


class MLPOptimizer(object):
    search_config: MLPSearchConfig
    best_trial_score: float
    best_config: Optional[MLPTrainingConfig]

    @staticmethod
    def stream(
        X: sps.csr_matrix, Y: sps.csc_matrix, mb_size: int, shuffle: bool = True
    ) -> Iterator[Tuple[jnp.ndarray, jnp.ndarray, int]]:
        assert X.shape[0] == Y.shape[0]
        shape_all: int = X.shape[0]
        index = np.arange(shape_all)
        if shuffle:
            np.random.shuffle(index)
        for start in range(0, shape_all, mb_size):
            end = min(shape_all, start + mb_size)
            mb_indices = index[start:end]
            yield (
                jnp.asarray(X[mb_indices].toarray(), dtype=jnp.float32),
                jnp.asarray(Y[mb_indices], dtype=jnp.float32),
                (end - start),
            )

    def __init__(
        self,
        profile: sps.csr_matrix,
        embedding: np.ndarray,
        search_config: Optional[MLPSearchConfig] = None,
    ):

        (
            profile_train,
            profile_test,
            embedding_train,
            embedding_test,
        ) = train_test_split(
            profile.astype(np.float32),
            embedding.astype(np.float32),
            random_state=42,
        )
        self.profile_train = profile_train
        self.profile_test = profile_test
        self.embedding_train = jnp.asarray(embedding_train, dtype=jnp.float32)
        self.embedding_test = jnp.asarray(embedding_test, dtype=jnp.float32)
        if search_config is None:
            self.search_config = MLPSearchConfig()
        else:
            self.search_config = search_config

    def search_best_config(
        self,
        n_trials: int = 10,
        logger: Optional[Logger] = None,
        random_seed: Optional[int] = None,
    ) -> Optional[MLPTrainingConfig]:
        self.best_trial_score = float("inf")
        self.best_config = None
        study = optuna.create_study(
            sampler=optuna.samplers.TPESampler(seed=random_seed)
        )
        if logger is None:
            logger = get_default_logger()
        r2 = (self.embedding_test ** 2).mean(axis=1).mean()
        logger.info("MSE baseline is %f", r2)

        def objective(trial: optuna.Trial) -> float:
            config = self.search_config.suggest(trial)
            mlp_function = hk.transform(
                lambda x, training: (
                    create_mlp(
                        self.embedding_train.shape[1],
                        config,
                    )
                )(x, training)
            )
            score, epoch = self._train_nn_with_trial(
                mlp_function, config=config, trial=trial
            )
            config.best_epoch = epoch
            if score < self.best_trial_score:
                self.best_trial_score = score
                self.best_config = config
            return score

        study.optimize(objective, n_trials=n_trials)
        return self.best_config

    def search_param_fit_all(
        self,
        n_trials: int = 10,
        logger: Optional[Logger] = None,
        random_seed: Optional[int] = None,
    ) -> Tuple[MLP, MLPTrainingConfig]:
        best_param = self.search_best_config(
            n_trials, logger=logger, random_seed=random_seed
        )

        if best_param is None:
            raise RuntimeError("An error occurred during the optimization step.")

        mlp = self.fit_full(best_param)
        return mlp, best_param

    def _train_nn_with_trial(
        self,
        mlp: hk.Transformed,
        config: MLPTrainingConfig,
        trial: Optional[optuna.Trial] = None,
    ) -> Tuple[float, int]:

        rng_key = jax.random.PRNGKey(0)
        rng_key, sub_key = jax.random.split(rng_key)
        params = mlp.init(
            sub_key,
            jnp.zeros((1, self.profile_train.shape[1]), dtype=jnp.float32),
            False,
        )
        opt = optax.adam(config.learning_rate)
        opt_state = opt.init(params)

        rng_key, sub_key = jax.random.split(rng_key)

        @partial(jax.jit, static_argnums=(3,))
        def predict(
            params: hk.Params, rng: PRNGKey, X: jnp.ndarray, training: bool
        ) -> jnp.ndarray:
            return mlp.apply(params, rng, X, training)

        @partial(jax.jit, static_argnums=(4,))
        def loss_fn(
            params: hk.Params,
            rng: PRNGKey,
            X: jnp.ndarray,
            Y: jnp.ndarray,
            training: bool,
        ) -> jnp.ndarray:
            prediction = predict(params, rng, X, training)
            return ((Y - prediction) ** 2).mean(axis=1).sum()

        @jax.jit
        def update(
            params: hk.Params,
            rng: PRNGKey,
            opt_state: optax.OptState,
            X: jnp.ndarray,
            Y: jnp.ndarray,
        ) -> Tuple[jnp.ndarray, hk.Params, optax.OptState]:
            loss_value = loss_fn(params, rng, X, Y, True)
            grad = jax.grad(loss_fn)(params, rng, X, Y, True)
            updates, opt_state = opt.update(grad, opt_state)
            new_params = optax.apply_updates(params, updates)
            return loss_value, new_params, opt_state

        best_val_score = float("inf")
        n_epochs = 512
        mb_size = 128
        score_degradation_count = 0
        val_score_degradation_max = 10
        best_epoch = 0
        for epoch in progress_bar(range(n_epochs)):
            train_loss = 0
            for X_mb, y_mb, _ in self.stream(
                self.profile_train, self.embedding_train, mb_size
            ):
                rng_key, sub_key = jax.random.split(rng_key)

                loss_value, params, opt_state = update(
                    params, sub_key, opt_state, X_mb, y_mb
                )
                train_loss += loss_value
            train_loss /= self.profile_train.shape[0]

            val_loss = 0
            for X_mb, y_mb, size in self.stream(
                self.profile_test, self.embedding_test, mb_size, shuffle=False
            ):
                val_loss += loss_fn(
                    params, rng_key, X_mb, y_mb, False
                )  # rng key will not be used
            val_loss /= self.profile_test.shape[0]
            if trial is not None:
                trial.report(val_loss, epoch)
                if trial.should_prune():
                    raise exceptions.TrialPruned()

            if val_loss < best_val_score:
                best_epoch = epoch + 1
                best_val_score = val_loss
                score_degradation_count = 0
            else:
                score_degradation_count += 1

            if score_degradation_count >= val_score_degradation_max:
                break

        return best_val_score, best_epoch

    def fit_full(self, config: MLPTrainingConfig) -> MLP:
        if config.best_epoch is None:
            raise ValueError("best epoch not specified by MLP Config")
        rng_key = jax.random.PRNGKey(0)

        mlp_function = hk.transform(
            lambda x, training: (
                create_mlp(
                    self.embedding_train.shape[1],
                    config,
                )
            )(x, training)
        )

        X = sps.vstack([self.profile_train, self.profile_test])
        y = jnp.concatenate([self.embedding_train, self.embedding_test], axis=0)
        mb_size = 128

        rng_key, sub_key = jax.random.split(rng_key)
        params = mlp_function.init(
            sub_key,
            jnp.zeros((1, self.profile_train.shape[1]), dtype=jnp.float32),
            True,
        )
        opt = optax.adam(config.learning_rate)
        opt_state = opt.init(params)

        @partial(jax.jit, static_argnums=(3,))
        def predict(
            params: hk.Params, rng: PRNGKey, X: jnp.ndarray, training: bool
        ) -> jnp.ndarray:
            return mlp_function.apply(params, rng, X, training)

        @partial(jax.jit, static_argnums=(4,))
        def loss_fn(
            params: hk.Params,
            rng: PRNGKey,
            X: jnp.ndarray,
            Y: jnp.ndarray,
            training: bool,
        ) -> jnp.ndarray:
            prediction = predict(params, rng, X, training)
            return ((Y - prediction) ** 2).mean(axis=1).sum()

        @jax.jit
        def update(
            params: hk.Params,
            rng: PRNGKey,
            opt_state: optax.OptState,
            X: jnp.ndarray,
            Y: jnp.ndarray,
        ) -> Tuple[jnp.ndarray, hk.Params, optax.OptState]:
            loss_value = loss_fn(params, rng, X, Y, True)
            grad = jax.grad(loss_fn)(params, rng, X, Y, True)
            updates, opt_state = opt.update(grad, opt_state)
            new_params = optax.apply_updates(params, updates)
            return loss_value, new_params, opt_state

        mb_size = 128
        for _ in progress_bar(range(config.best_epoch)):
            train_loss = 0
            for X_mb, y_mb, _ in self.stream(X, y, mb_size):
                rng_key, sub_key = jax.random.split(rng_key)
                loss_value, params, opt_state = update(
                    params, sub_key, opt_state, X_mb, y_mb
                )
                train_loss += loss_value
            train_loss /= self.profile_train.shape[0]
        return MLP(predict, params)



import os
import pickle
from contextlib import redirect_stderr, redirect_stdout
from logging import getLogger
from time import sleep
from typing import IO, Any, Dict

import numpy as np
import optuna
import pytest
import scipy.sparse as sps

from irspack import evaluator
from irspack.definitions import DenseScoreArray, InteractionMatrix, UserIndexArray
from irspack.optimizers.base_optimizer import BaseOptimizerWithEarlyStopping
from irspack.parameter_tuning import UniformSuggestion
from irspack.recommenders.base import BaseRecommender
from irspack.recommenders.base_earlystop import (
    BaseRecommenderWithEarlyStopping,
    TrainerBase,
)
from irspack.split import rowwise_train_test_split

X_small = sps.csr_matrix(
    (np.random.RandomState(42).rand(100, 32) > 0.5).astype(np.float64)
)
X_large = sps.csr_matrix(
    (np.random.RandomState(42).rand(50, 200) > 0.5).astype(np.float64)
)


class MockTrainer(TrainerBase):
    def __init__(self) -> None:
        self.epoch: int = 0

    def load_state(self, ifs: IO) -> None:
        self.epoch = pickle.load(ifs)["epoch"]

    def save_state(self, ofs: IO) -> None:
        pickle.dump(dict(epoch=self.epoch), ofs)

    def run_epoch(self) -> None:
        self.epoch += 1


class MockRecommender(BaseRecommenderWithEarlyStopping):
    trainer_class = MockTrainer

    def __init__(
        self,
        X: InteractionMatrix,
        target_epoch: int = 20,
        target_score: float = 0.0,
        **kwargs: Any,
    ):
        super().__init__(X, **kwargs)
        self.target_epoch = target_epoch
        self.target_score = target_score
        self.rns = np.random.RandomState(42)

    def _create_trainer(self) -> MockTrainer:
        sleep(0.01)
        return MockTrainer()

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        score = np.zeros(*self.X_train_all.shape)
        return score

    def _current_score(self) -> float:
        assert isinstance(self.trainer, MockTrainer)
        coeff: float
        if self.trainer.epoch > self.target_epoch:
            coeff = 1 - (self.trainer.epoch - self.target_epoch) / self.target_epoch
        else:
            coeff = self.trainer.epoch / self.target_epoch
        return self.target_score * coeff


class MockEvaluator(evaluator.Evaluator):
    def __init__(self, X: sps.csr_matrix) -> None:
        super().__init__(X, offset=0)
        self.target_metric = evaluator.TargetMetric.ndcg
        self.cutoff = 30

    def get_score(self, model: BaseRecommender) -> Dict[str, float]:
        assert isinstance(model, MockRecommender)
        return {self.target_metric.name: model._current_score()}


class MockOptimizer(BaseOptimizerWithEarlyStopping):
    recommender_class = MockRecommender
    default_tune_range = [UniformSuggestion("target_score", 0, 1)]


@pytest.mark.parametrize("X, target_epoch", [(X_small, 20)])
def test_optimizer_by_mock(X: InteractionMatrix, target_epoch: int) -> None:

    evaluator = MockEvaluator(X)
    optimizer = MockOptimizer(
        X,
        evaluator,
        fixed_params=dict(target_epoch=target_epoch),
        logger=getLogger("IGNORE"),
    )
    with redirect_stdout(open(os.devnull, "w")):
        with redirect_stderr(open(os.devnull, "w")):
            config, history = optimizer.optimize(n_trials=20, random_seed=42)
    assert len(config) == 2
    assert config["max_epoch"] == target_epoch
    best_index = np.nanargmax(-history.value.values)
    best_target_score_inferred = history.target_score.iloc[best_index]
    best_ndcg = history["ndcg@30"].iloc[best_index]

    assert best_target_score_inferred == pytest.approx(best_ndcg)
    assert config["target_score"] == pytest.approx(best_target_score_inferred)
    assert np.all(best_target_score_inferred >= history.target_score.values)



import os
import sys
from zipfile import ZipFile

import numpy as np
import pytest

from irspack.dataset.movielens import (
    MovieLens1MDataManager,
    MovieLens20MDataManager,
    MovieLens100KDataManager,
)

ZIPFILE_NAME = os.path.join(os.path.expanduser("~"), "ml.test.zip")


def test_ml100k() -> None:
    if sys.platform == "win32":
        pytest.skip("Skip on Windows.")
    GENRES = ["fantasy", "action", "thriller"]
    try:
        with ZipFile(ZIPFILE_NAME, "w") as zf:
            with zf.open("ml-100k/u.data", "w") as ofs:
                ofs.write("1\t2\t5\t0\n1\t3\t5\t86400".encode())

            with zf.open("ml-100k/u.genre", "w") as ofs:
                genre_string = ""
                for i, genre in enumerate(GENRES):
                    genre_string += f"{genre}|{i}\n"
                ofs.write(genre_string.encode())

            with zf.open("ml-100k/u.item", "w") as ofs:
                # movieId = 1 has action tag,
                # movieId = 2 has fantasy & thriller tags
                ofs.write(
                    """1|A fantastic movie|2020-01-01|2021-01-01|http://example.com|0|1|0
    2|Pandemic|2020-01-01|2021-01-01|http://example.com|1|0|1""".encode(
                        "latin-1"
                    )
                )
            with zf.open("ml-100k/u.user", "w") as ofs:
                ofs.write(
                    """1|32|M|0|1690074
    2|4|F|1|1760013
    """.encode()
                )

        loader = MovieLens100KDataManager(ZIPFILE_NAME)
        df = loader.read_interaction()
        movie_info, genres = loader.read_item_info()
        user_info = loader.read_user_info()
        assert df.shape == (2, 4)
        np.testing.assert_array_equal(df["userId"].values, [1, 1])
        np.testing.assert_array_equal(df["movieId"].values, [2, 3])
        np.testing.assert_array_equal(df["rating"].values, [5, 5])
        np.testing.assert_array_equal(
            df["timestamp"].values,
            np.asarray(
                [
                    "1970-01-01",
                    "1970-01-02",
                ],
                dtype="datetime64[ns]",
            ),
        )
        np.testing.assert_array_equal(movie_info.index.values, [1, 2])
        np.testing.assert_array_equal(
            movie_info.release_date,
            np.asarray(["2020-01-01", "2020-01-01"], dtype="datetime64[ns]"),
        )
        assert set(genres[genres.movieId == 1].genre) == set(["action"])
        assert set(genres[genres.movieId == 2].genre) == set(["fantasy", "thriller"])

        np.testing.assert_array_equal(user_info.index.values, [1, 2])
        np.testing.assert_array_equal(user_info.gender, ["M", "F"])
    finally:
        os.remove(ZIPFILE_NAME)


def test_ml1m() -> None:
    if sys.platform == "win32":
        pytest.skip("Skip on Windows.")

    try:
        with ZipFile(ZIPFILE_NAME, "w") as zf:
            with zf.open("ml-1m/ratings.dat", "w") as ofs:
                ofs.write(
                    """1::2::5::0
    1::3::5::86400
    """.encode()
                )
            with zf.open("ml-1m/movies.dat", "w") as ofs:
                ofs.write(
                    """1::A fantastic movie (2020)::fantasy|thriller
    1917::Vinni-Pukh(1969)::children
    """.encode(
                        "latin-1"
                    )
                )
            with zf.open("ml-1m/users.dat", "w") as ofs:
                ofs.write(
                    """1::M::32::0::1690074
    2::F::4::1::1760013
    """.encode()
                )

        loader = MovieLens1MDataManager(ZIPFILE_NAME)
        df = loader.read_interaction()
        movie_info = loader.read_item_info()
        user_info = loader.read_user_info()
        assert df.shape == (2, 4)
        np.testing.assert_array_equal(df["userId"].values, [1, 1])
        np.testing.assert_array_equal(df["movieId"].values, [2, 3])
        np.testing.assert_array_equal(df["rating"].values, [5, 5])
        np.testing.assert_array_equal(
            df["timestamp"].values,
            np.asarray(
                [
                    "1970-01-01",
                    "1970-01-02",
                ],
                dtype="datetime64[ns]",
            ),
        )
        np.testing.assert_array_equal(movie_info.index.values, [1, 1917])
        np.testing.assert_array_equal(movie_info.release_year, [2020, 1969])
        np.testing.assert_array_equal(user_info.index.values, [1, 2])
        np.testing.assert_array_equal(user_info.gender, ["M", "F"])
    finally:
        os.remove(ZIPFILE_NAME)


def test_ml20m() -> None:
    if sys.platform == "win32":
        pytest.skip("Skip on Windows.")
    try:
        with ZipFile(ZIPFILE_NAME, "w") as zf:
            with zf.open("ml-20m/ratings.csv", "w") as ofs:
                ofs.write(
                    """userId,movieId,rating,timestamp
    1,2,5,0
    1,3,5,86400
    """.encode()
                )
        loader = MovieLens20MDataManager(ZIPFILE_NAME)
        df = loader.read_interaction()
        np.testing.assert_array_equal(df["userId"].values, [1, 1])
        np.testing.assert_array_equal(df["movieId"].values, [2, 3])
        np.testing.assert_array_equal(df["rating"].values, [5, 5])
        np.testing.assert_array_equal(
            df["timestamp"].values,
            np.asarray(
                [
                    "1970-01-01",
                    "1970-01-02",
                ],
                dtype="datetime64[ns]",
            ),
        )
    finally:
        os.remove(ZIPFILE_NAME)



from .base import BaseEncoder
from .categorical import CategoricalValueEncoder
from .dataframe import DataFrameEncoder
from .float_categorizer import BinningEncoder
from .multi_value import ManyToManyEncoder

__all__ = [
    "BaseEncoder",
    "CategoricalValueEncoder",
    "BinningEncoder",
    "ManyToManyEncoder",
    "DataFrameEncoder",
]



from typing import Optional

from sklearn.decomposition import NMF

from ..definitions import DenseScoreArray, InteractionMatrix, UserIndexArray
from .base import BaseRecommender


class NMFRecommender(BaseRecommender):
    def __init__(
        self,
        X_train_all: InteractionMatrix,
        n_components: int = 64,
        alpha: float = 1e-2,
        l1_ratio: float = 1e-2,
        beta_loss: str = "frobenius",
        init: Optional[str] = None,
    ):
        super().__init__(X_train_all)
        self.n_components = n_components
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.beta_loss = beta_loss
        self.init = init

    def _learn(self) -> None:
        nmf_model = NMF(
            n_components=self.n_components,
            alpha=self.alpha,
            init=self.init,
            l1_ratio=self.l1_ratio,
            beta_loss=self.beta_loss,
            random_state=42,
        )
        self.nmf_model = nmf_model
        self.nmf_model.fit(self.X_train_all)
        self.W = self.nmf_model.fit_transform(self.X_train_all.tocsr())
        self.H = self.nmf_model.components_

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return self.W[user_indices].dot(self.H)

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        return self.nmf_model.transform(X).dot(self.H)



from collections import Counter
from typing import Generic, List, Optional, TypeVar, Union

import numpy as np
import pandas as pd
from scipy import sparse as sps

T = TypeVar("T")


class ManyToManyEncoder(Generic[T]):
    def __init__(self, items: List[T], normalize: bool = True, min_freq: int = 1):
        counter = Counter(items)
        self.unique_items = [item for item, cnt in counter.items() if cnt >= min_freq]
        self.names: List[Union[str, T]] = ["UNKNOWN"]
        self.names.extend(self.unique_items)
        self.normalize = normalize
        self.item_to_index = {item: i + 1 for i, item in enumerate(self.unique_items)}

    def __len__(self) -> int:
        return len(self.names)

    def transform_sparse(
        self,
        index: pd.DataFrame,
        items_df: pd.DataFrame,
        index_name_main: str,
        item_column_name: str,
        index_name_items: Optional[str] = None,
    ) -> sps.csr_matrix:
        unique_ids, inverse = np.unique(index[index_name_main], return_inverse=True)
        id_to_index = {id: i for i, id in enumerate(unique_ids)}
        relevant_items = items_df[items_df[index_name_items].isin(unique_ids)]
        row = relevant_items[index_name_items].map(id_to_index).values
        col = (
            relevant_items[item_column_name]
            .map(self.item_to_index)
            .fillna(0.0)
            .astype(np.int64)
            .values
        )
        result = sps.csr_matrix(
            (
                np.ones(len(row), dtype=np.float64),
                (row, col),
            ),
            shape=(len(unique_ids), len(self)),
        )
        result.sort_indices()
        if self.normalize:
            row_norm = result.power(2).sum(axis=1).A1 ** 0.5
            result.data /= row_norm[result.nonzero()[0]]
        return result[inverse]



from irspack.parameter_tuning.parameter_range import (
    CategoricalSuggestion,
    IntegerLogUniformSuggestion,
    IntegerSuggestion,
    LogUniformSuggestion,
    Suggestion,
    UniformSuggestion,
    is_valid_param_name,
    overwrite_suggestions,
)

__all__ = [
    "Suggestion",
    "UniformSuggestion",
    "LogUniformSuggestion",
    "IntegerSuggestion",
    "IntegerLogUniformSuggestion",
    "CategoricalSuggestion",
    "overwrite_suggestions",
    "is_valid_param_name",
]



import uuid
from typing import Any, List

import numpy as np
import pandas as pd
import pytest
import scipy.sparse as sps
from numpy.core.fromnumeric import choose

from irspack.split import holdout_specific_interactions

RNS = np.random.RandomState(0)
n_users = 1000
n_items = 512
df_size = 10000

user_ids = np.asarray([str(uuid.uuid1()) for _ in range(n_users)])
item_ids = np.asarray([str(uuid.uuid1()) for _ in range(n_items)])

df_master = pd.DataFrame(
    dict(
        user_id=RNS.choice(user_ids, size=df_size, replace=True),
        item_id=RNS.choice(item_ids, size=df_size, replace=True),
    )
).drop_duplicates()

df_master["timestamp"] = RNS.randint(-1000, 1000, size=df_master.shape[0])


def test_holdout_particular_item_interaction() -> None:
    df = df_master.copy()
    val_validatable_user_ratio = 0.3
    test_validatable_user_ratio = 0.3
    train_validatable_user_ratio = (
        1 - val_validatable_user_ratio - test_validatable_user_ratio
    )
    validatable_item_ids = np.random.choice(
        item_ids, size=len(item_ids) // 5, replace=False
    )
    indicator = np.where(df.item_id.isin(validatable_item_ids).values)
    item_id_reprod, dataset = holdout_specific_interactions(
        df,
        "user_id",
        "item_id",
        interaction_indicator=indicator,
        validatable_user_ratio_val=val_validatable_user_ratio,
        validatable_user_ratio_test=test_validatable_user_ratio,
    )
    np.testing.assert_array_equal(item_id_reprod, np.sort(item_ids))
    item_id_to_index = {iid: i for i, iid in enumerate(item_id_reprod)}
    validatable_index = np.sort(
        np.asarray([item_id_to_index[iid] for iid in validatable_item_ids])
    )
    train = dataset["train"]
    val = dataset["val"]
    test = dataset["test"]

    assert (
        df.shape[0]
        == train.X_all.count_nonzero()
        + val.X_all.count_nonzero()
        + test.X_all.count_nonzero()
    )

    n_validatable_users_in_train = (
        (train.X_all)[:, validatable_index].sum(axis=1).A1 >= 1
    ).sum()

    total_validatable_users = n_validatable_users_in_train + val.n_users + test.n_users

    assert train_validatable_user_ratio >= (
        (n_validatable_users_in_train - 1) / total_validatable_users
    )
    assert train_validatable_user_ratio <= (
        (n_validatable_users_in_train + 1) / total_validatable_users
    )
    assert val_validatable_user_ratio <= ((val.n_users + 1) / total_validatable_users)
    assert val_validatable_user_ratio >= ((val.n_users - 1) / total_validatable_users)
    assert test_validatable_user_ratio <= ((test.n_users + 1) / total_validatable_users)
    assert test_validatable_user_ratio >= ((test.n_users - 1) / total_validatable_users)

    assert (
        total_validatable_users
        == np.unique(df[df.item_id.isin(validatable_item_ids)].user_id).shape[0]
    )
    assert val.n_users > 0
    assert test.n_users > 0
    assert val.X_test is not None
    assert val.X_test.sum(axis=1).A1.min() >= 1
    assert test.X_test is not None
    assert test.X_test.sum(axis=1).A1.min() >= 1

    def X_to_df(X: sps.csr_matrix, uids: List[Any]) -> pd.DataFrame:
        rows, cols = X.nonzero()
        return pd.DataFrame(
            dict(user_id=[uids[row] for row in rows], item_id=item_id_reprod[cols])
        )

    assert np.all(X_to_df(val.X_test, val.user_ids).item_id.isin(validatable_item_ids))
    assert np.all(
        ~X_to_df(val.X_train, val.user_ids).item_id.isin(validatable_item_ids)
    )

    assert np.all(
        X_to_df(test.X_test, test.user_ids).item_id.isin(validatable_item_ids)
    )
    assert np.all(
        ~X_to_df(test.X_train, test.user_ids).item_id.isin(validatable_item_ids)
    )

    df_reproduced_all = pd.concat(
        [
            X_to_df(train.X_all, train.user_ids),
            X_to_df(val.X_all, val.user_ids),
            X_to_df(test.X_all, test.user_ids),
        ]
    )
    assert (
        df_reproduced_all.merge(df_master, on=["user_id", "item_id"]).shape[0]
        == df_master.shape[0]
    )


def test_raise() -> None:
    df = df_master.copy()
    TS_CUTPOINT = 100
    validatable_user_ratio_val = 0.6
    validatable_user_ratio_test = 0.5
    validatable_interactions = (df.timestamp >= 0).values
    with pytest.raises(ValueError):
        unique_item_ids, dataset = holdout_specific_interactions(
            df,
            "user_id",
            "item_id",
            validatable_interactions,
            validatable_user_ratio_val=validatable_user_ratio_val,
            validatable_user_ratio_test=validatable_user_ratio_test,
            random_seed=0,
        )


def test_holdout_future() -> None:
    df = df_master.copy()
    TS_CUTPOINT = 100
    validatable_interactions = (df.timestamp >= 0).values
    validatable_user_ratio_val = 0.5
    validatable_user_ratio_test = 0.5
    unique_item_ids, dataset = holdout_specific_interactions(
        df,
        "user_id",
        "item_id",
        validatable_interactions,
        validatable_user_ratio_val=validatable_user_ratio_val,
        validatable_user_ratio_test=validatable_user_ratio_test,
        random_seed=0,
    )
    train_users = dataset["train"]
    val_users = dataset["val"]
    test_users = dataset["test"]

    df_past = df[df.timestamp < 0]
    df_future = df[df.timestamp >= 0]
    users_past_only = np.unique(
        df_past[~df_past.user_id.isin(df_future.user_id)].user_id
    )
    np.testing.assert_array_equal(train_users.user_ids, users_past_only)

    def X_to_df(X: sps.csr_matrix, uids: np.ndarray) -> pd.DataFrame:
        rows, cols = X.nonzero()
        return pd.DataFrame(
            dict(user_id=[uids[row] for row in rows], item_id=unique_item_ids[cols])
        )

    for userset in [val_users, test_users]:
        # check all of the train_interactions are in the future
        train_interactions = X_to_df(userset.X_train, userset.user_ids)
        train_interactions_with_ts = train_interactions.merge(
            df_master, on=["user_id", "item_id"], how="inner"
        )
        assert train_interactions.shape[0] == train_interactions_with_ts.shape[0]
        assert np.all(train_interactions_with_ts.timestamp.values < 0)

        # check all of the test_interactions are in the future
        test_interactions = X_to_df(userset.X_test, userset.user_ids)
        test_interactions_with_ts = test_interactions.merge(
            df_master, on=["user_id", "item_id"], how="inner"
        )

        assert test_interactions.shape[0] == test_interactions_with_ts.shape[0]
        assert np.all(test_interactions_with_ts.timestamp.values >= 0)
